"""The :py:class:`Anacreon` class is a wrapper around 
:py:class:`~anacreonlib.anacreon_async_client.AnacreonAsyncClient` that handles 
basic functionality (such as state management and merging partial updates).
"""
import collections
import dataclasses
from typing import (
    Sequence,
    cast,
    Any,
    Callable,
    DefaultDict,
    Dict,
    List,
    Optional,
    SupportsFloat,
    Tuple,
    Type,
    Union,
)
from anacreonlib.types.request_datatypes import *
from anacreonlib.types.type_hints import BattleObjective
from anacreonlib.types.scenario_info_datatypes import (
    Category,
    ScenarioInfo,
    ScenarioInfoElement,
)
from anacreonlib.types.response_datatypes import *
from anacreonlib import utils
from anacreonlib.anacreon_async_client import AnacreonAsyncClient
import logging
import asyncio
import functools
from dataclasses import dataclass

__all__ = ("ProductionInfo", "MilitaryForceInfo", "Anacreon")

# Map from stateful request body to client method name and API request body type
_stateful_request_bodies: Dict[
    Type[SerializableDataclass], Tuple[str, Type[AnacreonApiRequest]]
] = {
    AbortAttack: ("abort_attack", AbortAttackRequest),
    Attack: ("attack", AttackRequest),
    AlterImprovement: ("destroy_improvement", AlterImprovementRequest),
    BuyItem: ("buy_item", BuyItemRequest),
    DeployFleet: ("deploy_fleet", DeployFleetRequest),
    DesignateWorld: ("designate_world", DesignateWorldRequest),
    DisbandFleet: ("disband_fleet", DisbandFleetRequest),
    LaunchJumpMissile: ("launch_lams", LaunchJumpMissileRequest),
    RenameObject: ("rename_object", RenameObjectRequest),
    SellFleet: ("sell_fleet", SellFleetRequest),
    SetFleetDestination: ("set_fleet_destination", SetFleetDestinationRequest),
    SetIndustryAlloc: ("set_industry_alloc", SetIndustryAllocRequest),
    SetProductAlloc: ("set_product_alloc", SetProductAllocRequest),
    SetTradeRoute: ("set_trade_route", SetTradeRouteRequest),
    StopTradeRoute: ("stop_trade_route", StopTradeRouteRequest),
    TransferFleet: ("transfer_fleet", TransferFleetRequest),
}


IdValueMapping = Union[Dict[int, int], List[int]]
"""Either a dict mapping from resource id to resource qty, or a list with resource id and qty interleaved"""


def _ensure_resources_list(resources: IdValueMapping) -> List[int]:
    """Ensure that an IdValueMapping is a list that can be passed directly to the Anacreon API"""
    if isinstance(resources, dict):
        resources_list = list()
        for item_id, item_count in resources.items():
            resources_list.append(item_id)
            resources_list.append(item_count)
    else:
        resources_list = resources
    return resources_list


@dataclasses.dataclass(eq=True)
class ProductionInfo:
    """This is a :py:func:`dataclasses.dataclass`"""

    #: Amount of resource that is stockpiled on the world
    available: float = 0

    #: Amount of resource that was consumed last watch
    consumed: float = 0

    #: Amount of resource that was exported last watch
    exported: float = 0

    #: Amount of resource that was imported last watch
    imported: float = 0

    #: Amount of resource that was produced last watch
    produced: float = 0

    #: Amount of resource that would have been consumed last watch if there were
    #: no resource shortages
    consumed_optimal: float = 0

    #: Amount of resource that would have been exported last watch if there were
    #: no resource shortages
    exported_optimal: float = 0

    #: Amount of resource that would have been imported last watch if there were
    #: no resource shortages
    imported_optimal: float = 0

    #: Amount of resource that would have been produced last watch if there were
    #: no resource shortages
    produced_optimal: float = 0

    def __add__(self: "ProductionInfo", other: "ProductionInfo") -> "ProductionInfo":
        """Add two :class:`ProductionInfo` instances together elementwise

        Returns:
            ProductionInfo: the elementwise sum
        """
        return ProductionInfo(
            available=self.available + other.available,
            consumed=self.consumed + other.consumed,
            imported=self.imported + other.imported,
            exported=self.exported + other.exported,
            produced=self.produced + other.produced,
            consumed_optimal=self.consumed_optimal + other.consumed_optimal,
            exported_optimal=self.exported_optimal + other.exported_optimal,
            imported_optimal=self.imported_optimal + other.imported_optimal,
            produced_optimal=self.produced_optimal + other.produced_optimal,
        )

    def __sub__(self: "ProductionInfo", other: "ProductionInfo") -> "ProductionInfo":
        """Subtract two :class:`ProductionInfo` instances together elementwise

        Returns:
            ProductionInfo: the elementwise difference
        """
        return ProductionInfo(
            available=self.available - other.available,
            consumed=self.consumed - other.consumed,
            imported=self.imported - other.imported,
            exported=self.exported - other.exported,
            produced=self.produced - other.produced,
            consumed_optimal=self.consumed_optimal - other.consumed_optimal,
            exported_optimal=self.exported_optimal - other.exported_optimal,
            imported_optimal=self.imported_optimal - other.imported_optimal,
            produced_optimal=self.produced_optimal - other.produced_optimal,
        )


@dataclasses.dataclass
class MilitaryForceInfo:
    """This is a :py:func:`dataclasses.dataclass`"""

    #: The amount of space forces on a world/in a fleet, as would be shown
    #: in the game UI
    space_forces: float

    #: The amount of ground forces on a world/in a fleet, as would be shown
    #: in the game UI
    ground_forces: float

    #: The amount of space forces that come from units that shoot missiles
    #: (as opposed to beam weapons)
    missile_forces: float

    #: The amount of space forces that come from units that are ships
    #: (as opposed to fixed point units or satellite units)
    maneuvering_unit_forces: float


class Anacreon:
    @classmethod
    async def log_in(cls, game_id: str, username: str, password: str) -> "Anacreon":
        """Authenticate with the Anacreon API using a Multiverse username and password

        Example:

            >>> import asyncio
            >>> async def main():
            ...     client = await Anacreon.log_in("8JNJ7FNZ", "username", "password")
            ...     # do stuff with client
            ...
            >>> asyncio.run(main())


        Args:
            game_id (str): The game ID of the game you intend to manipulate.
                You can find the game ID by looking at the URL when you are
                playing Anacreon in your browser. e.g if the URL was
                ``anacreon.kronosaur.com/trantor.hexm?gameID=8JNJ7FNZ``,
                the game ID would be ``8JNJ7FNZ``.

            username (str): The username to your Multiverse account
            password (str): The password to your Multiverse account

        Returns:
            Anacreon: An instance of the API client wrapper, which can be used
            to programmatically perform actions within the game
        """
        auth_req = AuthenticationRequest(username=username, password=password)
        client = AnacreonAsyncClient()
        auth_token = (await client.authenticate_user(auth_req)).auth_token
        return await cls.from_auth_token(game_id, auth_token, client=client)

    @classmethod
    async def from_auth_token(
        cls,
        game_id: str,
        auth_token: str,
        *,
        client: Optional[AnacreonAsyncClient] = None,
    ) -> "Anacreon":
        """Authenticate with the Anacreon API using an auth token

        You can get an auth token by looking at your cookies for
        `anacreon.kronosaur.com` in your browser's dev console.

        Args:
            game_id (str): The game ID of the game you intend to manipulate
            auth_token (str): A valid Multiverse auth token

        Returns:
            Anacreon: An instance of the API client wrapper, which can be used
            to programmatically perform actions within the game
        """
        client = client or AnacreonAsyncClient()
        game_info = await client.get_game_info(auth_token, game_id)

        auth_info: AnacreonApiRequest = AnacreonApiRequest(
            auth_token=auth_token,
            game_id=game_id,
            sovereign_id=game_info.user_info.sovereign_id,
        )

        ret = cls(auth_info, game_info, client=client)
        await ret.get_objects()
        return ret

    def __init__(
        self,
        auth_info: AnacreonApiRequest,
        game_info: ScenarioInfo,
        client: Optional[AnacreonAsyncClient] = None,
    ) -> None:
        """Construct an Anacreon instance directly

        Consider using one of the helper class methods like :func:`~anacreonlib.Anacreon.log_in` instead

        Args:
            auth_info (AnacreonApiRequest): A stub API request containing a valid auth token
            game_info (ScenarioInfo): Info about the game, such as all the traits that
            client (Optional[AnacreonAsyncClient], optional): The low-level API client to use in order to make HTTP
                requests. Defaults to None.
        """
        self._auth_info: AnacreonApiRequest = auth_info
        self._get_objects_event = asyncio.Event()
        self._state_updated_event = asyncio.Event()
        self._force_calculator = _MilitaryForceCalculator.from_game_info(game_info)

        self.logger = logging.getLogger(str(self.__class__.__name__))

        #: the low level API client that is used to make HTTP requests to the Anacreon API
        self.client: AnacreonAsyncClient = client or AnacreonAsyncClient()

        #: The scenario info for the game
        self.game_info: ScenarioInfo = game_info

        #: A mapping from trait/resource ID to :class:`ScenarioInfoElement`
        self.scenario_info_objects: Dict[int, ScenarioInfoElement] = {
            item_id: item
            for item in game_info.scenario_info
            if (item_id := item.id) is not None
        }

        #: A mapping from world/fleet ID to :class:`World` or :class:`Fleet` instance
        self.space_objects: Dict[int, Union[World, Fleet]] = dict()

        #: A mapping from siege ID to :class:`Siege` instances
        self.sieges: Dict[int, Siege] = dict()

        #: A mapping from sovereign ID to :class:`Sovereign` instance
        self.sovereigns: Dict[int, Sovereign] = {
            sov.id: sov for sov in game_info.sovereigns
        }

        #: A list of all history popups that would appear over your worlds in the game UI
        self.history: Dict[int, HistoryElement] = dict()

        self.update_obj: Optional[UpdateObject] = None

    @property
    def sov_id(self) -> int:
        """The sovereign ID of the currently logged in player"""
        return self._auth_info.sovereign_id

    async def get_objects(self) -> "Anacreon":
        """Refreshes game state from the Anacreon API to update world state,
        fleet state, and so on.

        Returns:
            Anacreon: this object
        """
        partial_state = await self.client.get_objects(self._auth_info)
        self._process_update(partial_state)

        # This is fine because the python event loop is single threaded
        # and there is no `await` in between the set/clear
        # so we know that nobody will call `wait_for_get_objects` in between
        # the .set() and the .clear() calls
        self._get_objects_event.set()
        self._get_objects_event.clear()

        return self

    async def wait_for_get_objects(self) -> None:
        """This coroutine waits until :meth:`Anacreon.get_objects` is called by
        someone else.

        One use case for this function is when you have 2 long-lived async
        :py:class:`~asyncio.Task` s running. One is caling
        :meth:`Anacreon.get_objects` every minute, while the other is
        managing a fleet. The task managing the fleet could to use this
        method to wait for updates to the state
        """
        await self._get_objects_event.wait()

    async def wait_for_any_update(self) -> None:
        """This coroutine waits until any method that updates game state, such as
        :func:`Anacreon.designate_world` or :func:`Anacreon.attack`, is called.

        One use case for this function is when you are just starting a script,
        and you spawn an :py:class:`~asyncio.Task` to periodically fetch updates.
        Your main coroutine would have to block until the task fetches the first
        update.
        """
        await self._state_updated_event.wait()

    def call_get_objects_periodically(self) -> "asyncio.Task[None]":
        """Spawns an :py:class:`~asyncio.Task` that calls
        :meth:`Anacreon.get_objects` every watch.

        Returns:
            asyncio.Task[None]: The task that was spawned.
        """

        async def _update() -> None:
            while True:
                await self.get_objects()
                if self.update_obj:
                    await asyncio.sleep(self.update_obj.next_update_time // 1000)
                else:
                    await asyncio.sleep(60)

        return asyncio.create_task(_update())

    def _process_update(
        self, partial_state: List[AnacreonObject]
    ) -> Optional[Selection]:
        selection = None

        for obj in partial_state:
            if isinstance(obj, (World, Fleet)):
                self.space_objects[obj.id] = obj

            elif isinstance(obj, BattlePlanObject):
                self.space_objects[obj.id].battle_plan = obj.battle_plan

            elif isinstance(obj, DestroyedSpaceObject):
                del self.space_objects[obj.id]

            elif isinstance(obj, Sovereign):
                self.sovereigns[obj.id] = obj

            elif isinstance(obj, Relationship):
                self.sovereigns[obj.id].relationship = obj.relationship

            elif isinstance(obj, Siege):
                self.sieges[obj.id] = obj

            elif isinstance(obj, UpdateObject):
                self._auth_info.sequence = obj.sequence
                self.update_obj = obj

            elif isinstance(obj, History):
                self.history = {h.id: h for h in obj.history}

            elif isinstance(obj, RegionObject):
                pass

            elif isinstance(obj, Selection):
                selection = obj

            elif isinstance(obj, dict):
                None

        self._state_updated_event.set()
        self._state_updated_event.clear()
        return selection

    async def _do_action(self, request: SerializableDataclass) -> Optional[Selection]:
        client_method_name, api_type = _stateful_request_bodies[type(request)]
        api_request = api_type(**request.dict(), **self._auth_info.dict())
        client_method = getattr(self.client, client_method_name)
        updated_objects: List[AnacreonObject] = await client_method(api_request)
        return self._process_update(updated_objects)

    # region: methods that call the api and update self.space_objects/related state

    async def abort_attack(self, battlefield_id: int) -> None:
        """Abort an attack in progress

        Args:
            battlefield_id (int): The world ID of the planet that you are attacking
        """
        await self._do_action(AbortAttack(battlefield_id=battlefield_id))

    async def attack(
        self,
        battlefield_id: int,
        objective: BattleObjective,
        enemy_sovereign_ids: List[int],
    ) -> None:
        """Attack a world

        Args:
            battlefield_id (int): The ID of the world that you intend to attack
            objective (BattleObjective): Whether you want to invade, reinforce a
                siege, or destroy all space defenses
            enemy_sovereign_ids (List[int]): A list of the sovereign ids that
                you wish to engage in battle with
        """
        await self._do_action(
            Attack(
                attacker_obj_id=battlefield_id,
                battle_plan=BattlePlan(
                    battlefield_id=battlefield_id,
                    objective=objective,
                    enemy_sovereign_ids=enemy_sovereign_ids,
                ),
            )
        )

    async def build_improvement(self, world_obj_id: int, improvement_id: int) -> None:
        """Build an improvement

        Args:
            world_obj_id (int): The ID of the world to build the improvement on
            improvement_id (int): The trait ID of the improvement to build
        """
        partial_update = await self.client.build_improvement(
            AlterImprovementRequest(
                source_obj_id=world_obj_id,
                improvement_id=improvement_id,
                **self._auth_info.dict(),
            )
        )
        self._process_update(partial_update)

    async def buy_item(self, source_obj_id: int, item_id: int, item_count: int) -> None:
        """Buy an item from the Mesophons.

        As a prerequisite to calling this method, you must have a fleet
        stationed at the world where you wish to buy ships from.

        Args:
            source_obj_id (int): The ID of the world to buy ships from
            item_id (int): The resource ID of the ship that you wish to buy
            item_count (int): The number of ships that you wish to buy. Unlike
                the UI, which only lets you buy ships in groups of 1000, you may
                specify any number here.
        """
        await self._do_action(
            BuyItem(source_obj_id=source_obj_id, item_id=item_id, item_count=item_count)
        )

    async def deploy_fleet(
        self, source_obj_id: int, resources: IdValueMapping
    ) -> Optional[Fleet]:
        """Deploy a fleet

        Args:
            source_obj_id (int): The world or fleet from which you wish to
                create the fleet
            resources (IdValueMapping): A mapping from resource IDs to quantities.
                Can either be a dict, or a flat list alternating between ID and
                qty.

        Returns:
            Optional[Fleet]: The newly created fleet.
        """
        selection = await self._do_action(
            DeployFleet(
                source_obj_id=source_obj_id, resources=_ensure_resources_list(resources)
            )
        )
        if selection is not None:
            ret = self.space_objects[selection.id]
            if isinstance(ret, Fleet):
                return ret
        return None

    async def designate_world(self, world_obj_id: int, designation_id: int) -> None:
        """Designate a world

        Args:
            world_obj_id (int): The ID of the world you wish to designate
            designation_id (int): The trait ID of the designation you wish to
                give the world
        """
        await self._do_action(
            DesignateWorld(source_obj_id=world_obj_id, new_designation=designation_id)
        )

    async def destroy_improvement(self, world_obj_id: int, improvement_id: int) -> None:
        """Destroy a structure/improvement on a world

        Args:
            world_obj_id (int): The ID of the world on which you want to destroy
                the structure
            improvement_id (int): The trait ID of the structure you wish to
                destroy
        """
        partial_update = await self.client.destroy_improvement(
            AlterImprovementRequest(
                source_obj_id=world_obj_id,
                improvement_id=improvement_id,
                **self._auth_info.dict(),
            )
        )
        self._process_update(partial_update)

    async def disband_fleet(self, fleet_obj_id: int, dest_obj_id: int) -> None:
        """Delete a fleet by forcing it to become a part of another object that
        it is stationed next to.

        You can use this method to gift fleets to enemy sovereigns.

        Args:
            fleet_obj_id (int): The ID of the fleet to disband
            dest_obj_id (int): The ID of the world/fleet that it should become a
                part of
        """
        await self._do_action(
            DisbandFleet(fleet_obj_id=fleet_obj_id, dest_obj_id=dest_obj_id)
        )

    async def launch_lams(self, source_obj_id: int, target_obj_id: int) -> None:
        """Launch jumpmissiles from a citadel

        Args:
            source_obj_id (int): The world ID of the citadel you wish to launch
                jumpmissiles from
            target_obj_id (int): The world/fleet ID you wish to shoot
                jumpmissiles at
        """
        await self._do_action(
            LaunchJumpMissile(source_obj_id=source_obj_id, target_obj_id=target_obj_id)
        )

    async def rename_object(self, obj_id: int, new_name: str) -> None:
        """Change the name of a world/fleet

        Args:
            obj_id (int): The ID of the object you wish to rename
            new_name (str): The object's new name
        """
        await self._do_action(RenameObject(obj_id=obj_id, name=new_name))

    async def sell_fleet(
        self, fleet_id: int, buyer_obj_id: int, resources: IdValueMapping
    ) -> None:
        """Sell a fleet (or fleet cargo) to the Mesophons

        As a prerequisite to calling this method, the fleet in question must
        already be stationed at a mesophon world.

        Args:
            fleet_id (int): The ID of the fleet to sell from
            buyer_obj_id (int): The ID of the Mesophon world you are selling to
            resources (IdValueMapping): The resources from the fleet which you
                are selling to the mesophons. You may choose to sell a portion
                of your resources, as opposed to all of the fleet's cargo or all
                of the ships.
        """
        await self._do_action(
            SellFleet(
                fleet_id=fleet_id,
                buyer_obj_id=buyer_obj_id,
                resources=_ensure_resources_list(resources),
            )
        )

    async def set_fleet_destination(self, fleet_obj_id: int, dest_obj_id: int) -> None:
        """Send a fleet somewhere

        Args:
            fleet_obj_id (int): The ID of the fleet
            dest_obj_id (int): The world ID of the fleet's desired destination
        """
        await self._do_action(
            SetFleetDestination(obj_id=fleet_obj_id, dest=dest_obj_id)
        )

    async def set_industry_alloc(
        self, world_id: int, industry_id: int, pct_labor_allocated: SupportsFloat
    ) -> None:
        """On a world, set the percent labor allocation to a structure

        For example, this method can be used to set the labor allocation to
        defense structures to 0%.

        Args:
            world_id (int): The ID of the world that the industry is on
            industry_id (int): The trait ID of the industry in question
            pct_labor_allocated (SupportsFloat): The percent of total labor on
                the world to allocate to this structure. This value should be
                between 0 and 100.
        """
        await self._do_action(
            SetIndustryAlloc(
                world_id=world_id,
                industry_id=industry_id,
                alloc_value=float(pct_labor_allocated),
            )
        )

    async def set_product_alloc(
        self,
        world_id: int,
        industry_id: int,
        pct_labor_allocated_by_resource: IdValueMapping,
    ) -> None:
        """Within a single industry, set the percent labor allocation to a
        single product of the industry.

        For example, this method can be used to set the labor allocation on a
        jumpship yards planet to spend 100% of the labor available to the yards
        structure on building Reliant-class jumptransports.

        Args:
            world_id (int): The ID of the world in question
            industry_id (int): The trait ID of the industry in question
            pct_labor_allocated_by_resource (IdValueMapping): A mapping from
                resource ID to percent labor allocation (from 0 to 100). This
                can either be a dict, or an alternating list.
        """
        await self._do_action(
            SetProductAlloc(
                world_id=world_id,
                industry_id=industry_id,
                alloc=_ensure_resources_list(pct_labor_allocated_by_resource),
            )
        )

    async def set_trade_route(
        self,
        importer_id: int,
        exporter_id: int,
        alloc_type: str = TradeRouteTypes.DEFAULT,
        alloc_value: Optional[Union[str, float]] = None,
        res_type_id: Optional[int] = None,
    ) -> None:
        """Create/update a trade route between 2 planets

        Args:
            importer_id (int): The ID of the world that is importing resources
            exporter_id (int): The ID of the world that is exporting resources
            alloc_type (TradeRouteTypes, optional): How to decide how the trade
                route should be set. Defaults to
                :py:attr:`TradeRouteTypes.DEFAULT`, which emulates the behavior
                of setting up a 1 way trade route between 2 unconnected planets
                in the game UI.
            alloc_value (Optional[Union[str, float]], optional): The percent of
                demand on the importing world to import from (between 0 and 999).
                Defaults to None.
            res_type_id (Optional[int], optional): The ID of the resource to
                import.
        """
        await self._do_action(
            SetTradeRoute(
                importer_id=importer_id,
                exporter_id=exporter_id,
                alloc_type=alloc_type,
                alloc_value=alloc_value,
                res_type_id=res_type_id,
            )
        )

    async def stop_trade_route(self, planet_id_a: int, planet_id_b: int) -> None:
        """Stop all trade between 2 planets

        If you need to stop trade on only one resource, you may do the set the
        percent of demand to import to 0 using :meth:`Anacreon.set_trade_route`
        with ``alloc_type`` as :py:attr:`TradeRouteTypes.CONSUMPTION`.

        Args:
            planet_id_a (int): The ID of one of the worlds on the trade route
            planet_id_b (int): The ID of another world on the trade route
        """
        await self._do_action(
            StopTradeRoute(planet_id_a=planet_id_a, planet_id_b=planet_id_b)
        )

    async def transfer_fleet(
        self, fleet_obj_id: int, dest_obj_id: int, resources: IdValueMapping
    ) -> None:
        """Transfer ships in a fleet to another fleet/world

        Args:
            fleet_obj_id (int): The ID of the fleet to transfer from
            dest_obj_id (int): The ID of the fleet/world to transfer to
            resources (IdValueMapping): The resources from the fleet to transfer
        """
        await self._do_action(
            TransferFleet(
                fleet_obj_id=fleet_obj_id,
                dest_obj_id=dest_obj_id,
                resources=_ensure_resources_list(resources),
            )
        )

    # endregion
    # region: methods that call the api but don't alter self.space_objects/related state

    async def get_tactical(self, battlefield_id: str) -> List[Dict[str, Any]]:
        """Get battlefield information of a planet, such as battle groups and squadron locations

        Args:
            battlefield_id (str): The ID of the world to check

        Returns:
            List[Dict[str, Any]]: A list of dicts that is essentially the raw JSON
            returned by the API.

        Todo:
            Add models for this method to make it more type safe
        """
        return await self.client.get_tactical(
            GetTacticalRequest(battlefield_id=battlefield_id, **self._auth_info.dict())
        )

    async def tactical_order(
        self,
        battlefield_id: int,
        order: Union[TacticalOrderType, str],
        squadron_id: int,
        orbit: Optional[float] = None,
        target_id: Optional[int] = None,
    ) -> bool:
        """Send an order to a tactial squadron on a world

        Args:
            battlefield_id (int): The ID of the world
            order (Union[TacticalOrderType, str]): The order you are giving
            squadron_id (int): The ID of the squadron to give the order to
            orbit (Optional[float]): If ``order`` is
                :attr:`TacticalOrderType.ORBIT`, this should be the new orbit
                altitude
            target_id (Optional[int]): If ``order`` is
                :attr:`TacticalOrderType.TARGET`, this should be the id of the
                tactical squadron to target

        Returns:
            bool: ``True`` if the order was successfully processed
        """
        return await self.client.tactical_order(
            TacticalOrderRequest(
                battlefield_id=battlefield_id,
                order=order,
                squadron_id=squadron_id,
                orbit=orbit,
                target_id=target_id,
                **self._auth_info.dict(),
            )
        )

    async def set_history_read(self, history_id: int) -> bool:
        """Delet a history popup that show up over a planet

        You can see a list of all the history popups by looking at
        :py:attr:`Anacreon.history`.

        Args:
            history_id (int): The history ID of the popup as per
                :py:attr:`HistoryElement.id`.

        Returns:
            bool: ``True`` if the popup was successfully closed.
        """
        successfully_cleared = await self.client.set_history_read(
            SetHistoryReadRequest(history_id=history_id, **self._auth_info.dict())
        )
        if successfully_cleared:
            del self.history[history_id]

        return successfully_cleared

    async def send_message(self, recipient_sov_id: int, message_text: str) -> None:
        """Send a message to another empire

        Args:
            recipient_sov_id (int): The sovereign ID of the recipient empire
            message_text (str): The text of the message
        """
        await self.client.send_message(
            SendMessageRequest(
                recipient_id=recipient_sov_id,
                message_text=message_text,
                **self._auth_info.dict(),
            )
        )

    # endregion
    # region: util methods

    def generate_production_info(
        self, world: Union[World, int]
    ) -> Dict[int, ProductionInfo]:
        """Calculate production info for a world

        Args:
            world (Union[World, int]): Either the :class:`World` object, or the
                world ID.

        Raises:
            LookupError: Raised if `world` is a world ID that cannot be found

        Returns:
            Dict[int, ProductionInfo]: A mapping from resource ID to
            :class:`ProductionInfo` objects describing how much of that
            resource was imported/exported
        """

        # This is more or less exactly how the game client calculates production as well
        # I ported this from JavaScript
        if isinstance(world, int):
            maybe_world_obj = self.space_objects[world]
            if maybe_world_obj is None or not isinstance(maybe_world_obj, World):
                raise LookupError(f"Could not find world with id {world}")
            worldobj: World = maybe_world_obj
        else:
            worldobj = world
        assert isinstance(worldobj, World)

        result: DefaultDict[int, ProductionInfo] = collections.defaultdict(
            ProductionInfo
        )

        flat_list_to_4tuples = cast(
            Callable[
                [Sequence[Union[int, float, None]]],
                List[Tuple[int, float, float, Optional[float]]],
            ],
            functools.partial(utils.flat_list_to_n_tuples, 4),
        )
        flat_list_to_3tuples = cast(
            Callable[
                [Sequence[Union[int, float, None]]],
                List[Tuple[int, float, Optional[float]]],
            ],
            functools.partial(utils.flat_list_to_n_tuples, 3),
        )

        resource_id: int
        optimal: float
        actual: Optional[float]

        if isinstance(worldobj, OwnedWorld):
            # First we take into account the base consumption of the planet (i.e the food the population eats)
            for resource_id, optimal, actual in flat_list_to_3tuples(
                worldobj.base_consumption
            ):
                entry = result[resource_id]

                entry.consumed_optimal += optimal

                if actual is None:
                    entry.consumed += optimal
                else:
                    entry.consumed += actual

        for trait in worldobj.traits:
            # Next, we take into account what our structures are consuming (i.e tril spent on growing food)
            if isinstance(trait, Trait):
                if trait.production_data:
                    for resource_id, optimal, actual in flat_list_to_3tuples(
                        trait.production_data
                    ):
                        entry = result[resource_id]

                        if optimal > 0.0:
                            entry.produced_optimal += optimal
                            if actual is None:
                                entry.produced += optimal
                            else:
                                entry.produced += actual
                        else:
                            entry.consumed_optimal += -optimal
                            if actual is None:
                                entry.consumed += -optimal
                            else:
                                entry.consumed += -actual

        if worldobj.trade_routes:
            # Finally, we account for trade routes
            for trade_route in worldobj.trade_routes:
                exports: Optional[List[Optional[float]]] = None
                imports: Optional[List[Optional[float]]] = None
                if trade_route.reciprocal:
                    # The data for this trade route belongs to another planet
                    partner_obj = self.space_objects.get(
                        trade_route.partner_obj_id, None
                    )
                    # would be sorta dumb if our trade route partner didn't actually exist
                    assert isinstance(
                        partner_obj, World
                    ), f"(world {worldobj.id}) partner id {repr(trade_route.partner_obj_id)} was a {type(partner_obj)} instead of World!"

                    # would also be dumb if our trade route partner didn't have any trade routes
                    assert (
                        partner_trade_routes := partner_obj.trade_route_partners
                    ) is not None

                    partner_trade_route = partner_trade_routes[worldobj.id]
                    imports = partner_trade_route.exports
                    exports = partner_trade_route.imports
                else:
                    exports = trade_route.exports
                    imports = trade_route.imports

                if exports is not None:
                    for resource_id, _pct, optimal, actual in flat_list_to_4tuples(
                        exports
                    ):
                        entry = result[resource_id]

                        if actual is None:
                            entry.exported += optimal
                        else:
                            entry.exported += actual

                        entry.exported_optimal += optimal

                if imports is not None:
                    for resource_id, _pct, optimal, actual in flat_list_to_4tuples(
                        imports
                    ):
                        entry = result[resource_id]

                        if actual is None:
                            entry.imported += optimal
                        else:
                            entry.imported += actual

                        entry.imported_optimal += optimal

                if worldobj.resources:
                    for resource_id, resource_qty in cast(
                        List[Tuple[int, float]],
                        utils.flat_list_to_n_tuples(2, worldobj.resources),
                    ):
                        if resource_qty > 0:
                            result[resource_id].available = resource_qty

        return {int(k): v for k, v in result.items()}

    def calculate_forces(
        self, object_or_resources: Union[World, Fleet, IdValueMapping]
    ) -> MilitaryForceInfo:
        """Calculate the ground forces + space forces of a particular world/fleet

        Args:
            object_or_resources (Union[World, Fleet, IdValueMapping]): Either
                the :class:`World` object, the :class:`Fleet` object, or a
                list/dict of resources.

        Returns:
            MilitaryForceInfo: A dataclass containing the force information as
            it would be displayed in the Anacreon UI
        """
        return self._force_calculator.calculate_forces(object_or_resources)

    def calculate_remaining_cargo_space(self, fleet: Union[Fleet, int]) -> float:
        """Calculate the remaining cargo space on a fleet

        Args:
            fleet (Union[Fleet, int]): Either the :class:`Fleet` object, or the
                fleet ID

        Raises:
            LookupError: Raised if the fleet could not be found

        Returns:
            float: The remaining cargo space left in the fleet. Can be negative
            if uneven attrition has left more cargo in the fleet than it has
            space for.
        """
        if isinstance(fleet, int):
            maybe_fleet = self.space_objects[fleet]
            if isinstance(maybe_fleet, Fleet):
                fleet = maybe_fleet
            else:
                raise LookupError(f"Could not find fleet with id {fleet}")

        fleet_resources = cast(
            List[Tuple[int, float]], utils.flat_list_to_tuples(fleet.resources)
        )

        remaining_cargo_space: float = 0
        for res_id, qty in fleet_resources:
            res_info = self.scenario_info_objects[res_id]
            if res_info.cargo_space:
                remaining_cargo_space += res_info.cargo_space * qty
            elif res_info.is_cargo and res_info.mass:
                remaining_cargo_space -= res_info.mass * qty

        return remaining_cargo_space

    def get_valid_improvement_list(self, world: World) -> List[ScenarioInfoElement]:
        """Returns a list of scenario elements which represent improvements that
        can be built on a given world.

        Args:
            world (World): The world in question

        Returns:
            List[ScenarioInfoElement]: A list of improvements that can be built.
        """
        valid_improvement_ids: List[ScenarioInfoElement] = []
        trait_dict = world.squashed_trait_dict

        # func returns true if this world has trait
        this_world_has_trait: Callable[[int], bool] = functools.partial(
            utils.world_has_trait, self.game_info.scenario_info, world
        )

        for improvement in self.game_info.scenario_info:
            if (
                improvement.category == Category.IMPROVEMENT  # should be an improvement
                and improvement.id is not None
                and improvement.id
                not in trait_dict.keys()  #  that is not already built
                and improvement.build_time is not None  #       that could be built
                and not improvement.npe_only  #                 by players
                and not improvement.designation_only  #         without redesignating
                and (
                    improvement.min_tech_level is None  #
                    or world.tech_level >= improvement.min_tech_level
                )
            ):
                if improvement.build_upgrade:
                    # Check if we have the predecessor structure.
                    has_predecessor_structure = any(
                        this_world_has_trait(predecessor)
                        and not utils.trait_under_construction(trait_dict, predecessor)
                        for predecessor in improvement.build_upgrade
                    )
                    if not has_predecessor_structure:
                        continue
                if improvement.build_requirements:
                    # Check we have requirements. Requirements can be any trait.
                    requirement_missing = any(
                        not this_world_has_trait(requirement_id)
                        or utils.trait_under_construction(trait_dict, requirement_id)
                        for requirement_id in improvement.build_requirements
                    )
                    if requirement_missing:
                        continue

                if improvement.build_exclusions:
                    # Check if we are banned from doing so
                    if any(
                        this_world_has_trait(exclusion_id)
                        for exclusion_id in improvement.build_exclusions
                    ):
                        continue

                # Check if this trait would be a downgrade from an existing trait
                if any(
                    utils.does_trait_depend_on_trait(
                        self.game_info.scenario_info, existing_trait_id, improvement.id
                    )
                    for existing_trait_id in trait_dict.keys()
                ):
                    continue

                # if this is a tech advancement structure, check if we can build it
                if improvement.role == "techAdvance":
                    if (improvement.tech_level_advance or 0) <= world.tech_level:
                        continue

                # we have not continue'd so it is ok to  build
                valid_improvement_ids.append(improvement)

        return valid_improvement_ids

    # endregion


_missile_unit_unids = {
    "core.GDM",
    "core.hypersonicMissile",
    "core.armoredSatellite",
    "core.battlestationTitan",
    "core.jumpcruiserAdamant",
    "core.jumpcruiserUndine",
    "core.starcruiserBehemoth",
    "core.starcruiserMegathere",
    "core.starcruiserTyphon",
    "core.starcruiserVictory",
}


@dataclass
class _MilitaryForceCalculator:
    sf_calc: Dict[int, float]
    gf_calc: Dict[int, float]
    maneuvering_unit_calc: Dict[int, float]
    missile_calc: Dict[int, float]

    @classmethod
    def from_game_info(cls, game_info: ScenarioInfo) -> "_MilitaryForceCalculator":
        """
        Generate the dictionaries required to calculate space and ground force of an object

        :return: None
        """

        sf_calc: Dict[int, float] = dict()
        gf_calc: Dict[int, float] = dict()
        maneuvering_unit_calc: Dict[int, float] = dict()
        missile_calc: Dict[int, float] = dict()

        for item in game_info.scenario_info:
            if item.attack_value is not None:
                assert item.id is not None
                attack_value = float(item.attack_value)

                if (
                    item.category
                    in (
                        Category.FIXED_UNIT,
                        Category.ORBITAL_UNIT,
                        Category.MANEUVERING_UNIT,
                    )
                    and item.cargo_space is None
                ):
                    sf_calc[item.id] = attack_value
                    if item.category == Category.MANEUVERING_UNIT:
                        maneuvering_unit_calc[item.id] = attack_value
                    if item.unid in _missile_unit_unids:
                        missile_calc[item.id] = attack_value

                elif item.category == Category.GROUND_UNIT:
                    gf_calc[item.id] = attack_value

        return cls(sf_calc, gf_calc, maneuvering_unit_calc, missile_calc)

    def calculate_forces(
        self, object_or_resources: Union[World, Fleet, IdValueMapping]
    ) -> MilitaryForceInfo:
        if isinstance(object_or_resources, (World, Fleet)):
            if object_or_resources.resources is None:
                return MilitaryForceInfo(0, 0, 0, 0)

            resource_list = object_or_resources.resources
        else:
            resource_list = _ensure_resources_list(object_or_resources)

        space_forces = 0.0
        ground_forces = 0.0
        maneuveringunit_force = 0.0
        missile_force = 0.0

        for item_id, item_qty in cast(
            List[Tuple[int, float]], utils.flat_list_to_tuples(resource_list)
        ):
            space_forces += float(item_qty) * self.sf_calc.get(item_id, 0)
            ground_forces += float(item_qty) * self.gf_calc.get(item_id, 0)
            maneuveringunit_force += float(item_qty) * self.maneuvering_unit_calc.get(
                item_id, 0
            )
            missile_force += float(item_qty) * self.missile_calc.get(item_id, 0)

        return MilitaryForceInfo(
            space_forces / 100,
            ground_forces / 100,
            maneuveringunit_force / 100,
            missile_force / 100,
        )



import json
from typing import Any, List, TypedDict
import unittest
from anacreonlib.types import response_datatypes
from pathlib import Path
from collections import Counter
from pprint import pformat

current_folder_path = Path(__file__).resolve().parent


class DeserializationTest(unittest.TestCase):
    def test_get_objects_load(self) -> None:
        class GetObjectTestCase(TypedDict):
            file_path: Path
            expected_owned_world_count: int

        paths = (
            GetObjectTestCase(
                file_path=current_folder_path / "getObjects_2020_07_18.json",
                expected_owned_world_count=379,
            ),
            GetObjectTestCase(
                file_path=current_folder_path / "getObjects_2020_08_22.json",
                expected_owned_world_count=782,
            ),
        )

        for case in paths:
            with self.subTest(f"test load of {case!r}"):
                self.check_get_objects_load_is_correct(**case)

    def check_get_objects_load_is_correct(
        self, file_path: Path, expected_owned_world_count: int
    ) -> None:
        with open(file_path, "r") as f:
            get_objects_json_response: List[Any] = json.load(f)

        deserialized_objects: List[response_datatypes.AnacreonObject] = [
            response_datatypes._convert_json_to_anacreon_obj(
                response_datatypes.AnacreonObject, obj
            )
            for obj in get_objects_json_response
        ]
        # All objects should be deserialized
        self.assertEqual(len(get_objects_json_response), len(deserialized_objects))

        # Count the types of objects we deserialised
        counter = Counter(map(type, deserialized_objects))

        # When I made the getObjects call, my empire had 379 worlds
        self.assertEqual(
            counter[response_datatypes.OwnedWorld], expected_owned_world_count
        )
        self.assertEqual(counter[response_datatypes.OwnSovereign], 1)

        # All of the objects should have been deserialized successfully
        # i.e none should still be in raw json format
        self.assertEqual(
            counter[dict],
            0,
            msg=f"""
Encountered a `dict` that did not get deserialized into pydantic
{pformat(v) if (v := next((v for v in deserialized_objects if isinstance(v, dict)), None)) else "wait actually no something is really gone wrong"}
""",
        )
        self.assertEqual(counter[list], 0)

        reserialized_json: List[Any] = [
            json.loads(obj.model_dump_json(by_alias=True))
            for obj in deserialized_objects
        ]
        self.assertEqual(len(get_objects_json_response), len(reserialized_json))

        for before, after in zip(get_objects_json_response, reserialized_json):
            # Check that `after` is a superset of `before`
            # Only a shallow check that all keys in before are in after - does not verify that such information was actually kept
            self.assertEqual(
                {**before, **after},
                after,
                msg=f"""
before: 
{pformat(before)}

------

after:
{pformat(after)}

------

keys in before that are not in after:
{pformat(before.keys() - after.keys())}
                """,
            )

    def test_aeon_ip_integer(self) -> None:
        raw = {
            "fleets": ["AEON2011:ipInteger:v1", "KzFQSQUAAAABOYqakg=="],
            "population": 10,
            "resources": [
                ["AEON2011:ipInteger:v1", "KzFQSQUAAAABOYqakg=="],
                ["AEON2011:ipInteger:v1", "KzFQSQUAAAABOYqakg=="],
            ],
            "techLevel": 9,
            "worlds": 10,
        }

        parsed = response_datatypes.SovereignStats.parse_obj(raw)

        self.assertEqual(5260352146, parsed.fleets)
        self.assertEqual([5260352146, 5260352146], parsed.resources)

    def test_owned_world_deserialization(self) -> None:
        # given: an API response from the player playing as the below sovereign id
        sov_id = 3913334
        with open(current_folder_path / "getObjects_2020_08_22.json", "r") as f:
            raw = json.load(f)

        # when: we parse all of the objects
        parsed_objects: List[response_datatypes.AnacreonObject] = [
            response_datatypes._convert_json_to_anacreon_obj(
                response_datatypes.AnacreonObject, obj
            )
            for obj in raw
        ]

        # then: all of the ones with class: world should be a world object, and
        # none of the ones without class: world should be a world object
        world_indexes = set(
            i for i, obj in enumerate(raw) if obj.get("class", None) == "world"
        )
        for i, obj in enumerate(parsed_objects):
            if i in world_indexes:
                try:
                    self.assertIsInstance(obj, response_datatypes.World)
                except AssertionError:
                    response_datatypes.AnacreonObject.model_validate(obj)
                    raise
            else:
                self.assertNotIsInstance(obj, response_datatypes.World)

        # when: we look at all of the world objects
        worlds = [w for w in parsed_objects if isinstance(w, response_datatypes.World)]

        # then: it should only be an owned world iff we actually own it
        for world in worlds:
            with self.subTest(f"test world id {world.id}"):
                if world.sovereign_id == sov_id:
                    self.assertIsInstance(world, response_datatypes.OwnedWorld)
                else:
                    self.assertNotIsInstance(world, response_datatypes.OwnedWorld)


if __name__ == "__main__":
    unittest.main()



# type: ignore
# uplink does not play well with type checking
import asyncio
from typing import List, Dict, Any

import aiohttp
import uplink
from uplink import Consumer, post, returns, Query, get, Body, clients
from uplink.retry import retry
from uplink.retry.when import raises

from anacreonlib.types.request_datatypes import *
from anacreonlib.types.response_datatypes import (
    AuthenticationResponse,
    convert_json_to_anacreon_obj,
    AnacreonObject,
    handle_hexarc_error_response,
)
from anacreonlib.types.scenario_info_datatypes import (
    ScenarioInfo,
    convert_json_to_scenario_info,
)


@uplink.timeout(10)
@uplink.retry(when=raises(retry.CONNECTION_TIMEOUT))
@uplink.json
@returns.json
@handle_hexarc_error_response
class AnacreonAsyncClient(Consumer):
    """
    A coroutine-based asynchronous API client to interact with anacreon
    """

    def __init__(
        self, *, base_url: str = "https://anacreon.kronosaur.com/api/"
    ) -> None:
        self._aio_session = aiohttp.ClientSession()
        super().__init__(
            base_url=base_url,
            client=clients.AiohttpClient(session=self._aio_session),
            converter=(
                pydantic_request_converter,
                convert_json_to_anacreon_obj,
                convert_json_to_scenario_info,
            ),
        )

    def __del__(self):
        """Ensures the session was closed"""
        if not self._aio_session.closed:
            asyncio.run(self._aio_session.close())

    @post("login")
    async def authenticate_user(
        self, username_and_pw: Body(type=AuthenticationRequest)
    ) -> AuthenticationResponse:
        """
        Logs you into Anacreon. Does not on its own throw an error if you get your password
        wrong!

        :type username: str
        :param username: Username of account to log in as

        :type password: str
        :param password: Password of account to log in as

        :param actual: If false, forces the request to fail
        :type actual: bool

        :return: JSON response.
        """

    @get("gameList")
    async def get_game_list(self, auth_token: Query("authToken")):
        """
        Get the list of games we are in right now (?)
        :return: said list
        """

    @get("getGameInfo")
    async def get_game_info(
        self, auth_token: Query("authToken"), game_id: Query("gameID")
    ) -> ScenarioInfo:
        """
        Get information about the game such as

        - Info about the scenario
            - All the items/designations/etc that could exist and their ID's
        - Info about the game
            - All the sovereigns and their ID's
        - Your user info
            - Your sovereign ID, capital ID, etc

        :return: Said information
        """

    @post("getObjects/")
    async def get_objects(
        self, request: Body(type=AnacreonApiRequest)
    ) -> List[AnacreonObject]:
        """
        :return: A list of all objects that you have explored and data relevant to them, such as object ID's, planet
        designations, resources contained in fleets, and similar information relevant to gameplay
        """

    @post("deployFleet")
    async def deploy_fleet(
        self, request: Body(type=DeployFleetRequest)
    ) -> List[AnacreonObject]:
        """
        Deploy a fleet
        :return: A refreshed version of the ``get_objects`` response
        """

    @post("transferFleet")
    async def transfer_fleet(
        self, request: Body(type=TransferFleetRequest)
    ) -> List[AnacreonObject]:
        """
        Transfer a fleet's resources

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("disbandFleet")
    async def disband_fleet(
        self, request: Body(type=DisbandFleetRequest)
    ) -> List[AnacreonObject]:
        """
        Disband a fleet to anyone/anything else

        Disbanding a fleet completely dissolves it. You can disband a fleet to yourself OR another sovereign.

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("renameObject")
    async def rename_object(
        self, request: Body(type=RenameObjectRequest)
    ) -> List[AnacreonObject]:
        """
        Rename an object that belongs to your sovereign

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("setDestination")
    async def set_fleet_destination(
        self, request: Body(type=SetFleetDestinationRequest)
    ) -> List[AnacreonObject]:
        """
        Send a fleet somewhere

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("attack")
    async def attack(self, request: Body(type=AttackRequest)) -> List[AnacreonObject]:
        """
        Initiate an attack on an object

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("abortAttack")
    async def abort_attack(
        self, request: Body(type=AbortAttackRequest)
    ) -> List[AnacreonObject]:
        """
        Abort an attack

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("launchLAMs")
    async def launch_lams(
        self, request: Body(type=LaunchJumpMissileRequest)
    ) -> List[AnacreonObject]:
        """
        Launch some jumpmissiles at a fleet

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("designateWorld")
    async def designate_world(
        self, request: Body(type=DesignateWorldRequest)
    ) -> List[AnacreonObject]:
        """
        Designate a world to something

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("buildImprovement")
    async def build_improvement(
        self, request: Body(type=AlterImprovementRequest)
    ) -> List[AnacreonObject]:
        """
        Build an improvement on a world

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("destroyImprovement")
    async def destroy_improvement(
        self, request: Body(type=AlterImprovementRequest)
    ) -> List[AnacreonObject]:
        """
        Destroy an improvement

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("setIndustryAlloc")
    async def set_industry_alloc(
        self, request: Body(type=SetIndustryAllocRequest)
    ) -> List[AnacreonObject]:
        """
        Change the allocation of an industry as a percent of labor on the world

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("setProductAlloc")
    async def set_product_alloc(
        self, request: Body(type=SetProductAllocRequest)
    ) -> List[AnacreonObject]:
        """
        Change the allocation of how a structure produces its products

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("setTradeRoute")
    async def set_trade_route(
        self, request: Body(type=SetTradeRouteRequest)
    ) -> List[AnacreonObject]:
        """
        Add a trade route between two worlds

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("stopTradeRoute")
    async def stop_trade_route(
        self, request: Body(type=StopTradeRouteRequest)
    ) -> List[AnacreonObject]:
        """
        Stop a trade route between two planets

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("buyItem")
    async def buy_item(
        self, request: Body(type=BuyItemRequest)
    ) -> List[AnacreonObject]:
        """
        Buy something

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("sellFleet")
    async def sell_fleet(
        self, request: Body(type=SellFleetRequest)
    ) -> List[AnacreonObject]:
        """
        Sell a fleet to a planet

        :return: A refreshed version of ``Anacreon.get_objects()``
        """

    @post("getTactical")
    async def get_tactical(
        self, battlefield_id: Body(type=GetTacticalRequest)
    ) -> List[Dict[str, Any]]:
        """
        Get battlefield information of a planet, such as battle groups and squadron locations

        :return: Battlefield info
        """

    @post("tacticalOrder")
    async def tactical_order(self, order: Body(type=TacticalOrderRequest)) -> bool:
        """
        Give a tactical order

        :return: If your order was carried out
        """

    @post("setHistoryRead")
    async def set_history_read(
        self, history_id: Body(type=SetHistoryReadRequest)
    ) -> bool:
        """
        Delete one of those popups that show up over planets

        :return: If the popup was cleared successfully
        """

    @post("sendMessage")
    async def send_message(self, message: Body(type=SendMessageRequest)) -> None:
        """
        Send a message to another empire

        :return: None
        """



from anacreonlib.types.response_datatypes import AnacreonObject, AuthenticationResponse
from anacreonlib.types.scenario_info_datatypes import ScenarioInfo
from typing import Any, Dict, List
from anacreonlib.types.request_datatypes import (
    AbortAttackRequest,
    AlterImprovementRequest,
    AnacreonApiRequest,
    AttackRequest,
    AuthenticationRequest,
    BuyItemRequest,
    DeployFleetRequest,
    DesignateWorldRequest,
    DisbandFleetRequest,
    GetTacticalRequest,
    LaunchJumpMissileRequest,
    RenameObjectRequest,
    SellFleetRequest,
    SendMessageRequest,
    SetFleetDestinationRequest,
    SetHistoryReadRequest,
    SetIndustryAllocRequest,
    SetProductAllocRequest,
    SetTradeRouteRequest,
    StopTradeRouteRequest,
    TacticalOrderRequest,
    TransferFleetRequest,
)

class AnacreonAsyncClient:
    """
    A coroutine-based asynchronous API client to interact with anacreon
    """

    def __init__(self, *, base_url: str = ...) -> None: ...
    async def authenticate_user(
        self, username_and_pw: AuthenticationRequest
    ) -> AuthenticationResponse:
        """Retrieve an auth token from the Anacreon API using Multiverse login
        credentials

        Args:
            username_and_pw (AuthenticationRequest): An object containing
                Multiverse login credentials

        Returns:
            AuthenticationResponse: Response object containing auth token
        """
    async def get_game_list(self, auth_token: str) -> List[Dict[Any, Any]]:
        """Get the list of games we are in right now

        Args:
            auth_token (str): API auth token

        Returns:
            List[Dict[Any, Any]]: JSON response
        """
    async def get_game_info(self, auth_token: str, game_id: str) -> ScenarioInfo:
        """Get information about the game

        This includes
        - Info about the scenario
            - All the items/designations/etc that could exist and their ID's
        - Info about the game
            - All the sovereigns and their ID's
        - Your user info
            - Your sovereign ID, capital ID, etc

        Args:
            auth_token (str): API auth token
            game_id (str): Which game to return info for

        Returns:
            ScenarioInfo: The game info
        """
    async def get_objects(self, request: AnacreonApiRequest) -> List[AnacreonObject]:
        """Get a list of all of the objects in the game

        Args:
            request (AnacreonApiRequest): Request object containing API auth token

        Returns:
            List[AnacreonObject]: A list of all objects in the game
        """
    async def deploy_fleet(self, request: DeployFleetRequest) -> List[AnacreonObject]:
        """Deploy a fleet

        Args:
            request (DeployFleetRequest): Request object containing API auth
                token, where to deploy the fleet, etc

        Returns:
            List[AnacreonObject]: A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def transfer_fleet(
        self, request: TransferFleetRequest
    ) -> List[AnacreonObject]:
        """Transfer a fleet's resources

        Args:
            request (TransferFleetRequest): Request object containing API auth
                token and details about the transfer

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def disband_fleet(self, request: DisbandFleetRequest) -> List[AnacreonObject]:
        """Disband a fleet, merging it with another object (even if the other
        object belongs to another sovereign)

        Args:
            request (DisbandFleetRequest): Request object containing API auth
                token and the ID of the fleet to disband

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def rename_object(self, request: RenameObjectRequest) -> List[AnacreonObject]:
        """Rename an object that belongs to your sovereign

        Args:
            request (RenameObjectRequest): Request object containing API auth
                token, details about the renaming operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def set_fleet_destination(
        self, request: SetFleetDestinationRequest
    ) -> List[AnacreonObject]:
        """Send a fleet somewhere

        Args:
            request (SetFleetDestinationRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def attack(self, request: AttackRequest) -> List[AnacreonObject]:
        """Initiate an attack on an object

        Args:
            request (AttackRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def abort_attack(self, request: AbortAttackRequest) -> List[AnacreonObject]:
        """Abort an attack

        Args:
            request (AbortAttackRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def launch_lams(
        self, request: LaunchJumpMissileRequest
    ) -> List[AnacreonObject]:
        """Launch some jumpmissiles at a fleet

        Args:
            request (LaunchJumpMissileRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def designate_world(
        self, request: DesignateWorldRequest
    ) -> List[AnacreonObject]:
        """Change a world's designation

        Args:
            request (DesignateWorldRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def build_improvement(
        self, request: AlterImprovementRequest
    ) -> List[AnacreonObject]:
        """Build an improvement on a world

        Args:
            request (AlterImprovementRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def destroy_improvement(
        self, request: AlterImprovementRequest
    ) -> List[AnacreonObject]:
        """Destroy an improvement

        Args:
            request (AlterImprovementRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def set_industry_alloc(
        self, request: SetIndustryAllocRequest
    ) -> List[AnacreonObject]:
        """For a single industry on a single world, change its percent
        allocation of total labor.

        Typically, you can only freely set the labor allocation of defense
        structures.

        Args:
            request (SetIndustryAllocRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def set_product_alloc(
        self, request: SetProductAllocRequest
    ) -> List[AnacreonObject]:
        """For a single product type produced by a single industry on a single
        world, change the percent allocation of labor in the industry allocated
        towards making that product.

        Typically, you can only freely set product labor allocation on shipyard
        industries (i.e you can tell your ship yards to produce 50% transports
        and 50% gunships)

        Args:
            request (SetProductAllocRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def set_trade_route(
        self, request: SetTradeRouteRequest
    ) -> List[AnacreonObject]:
        """Add or update a trade route between two worlds

        Args:
            request (SetTradeRouteRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def stop_trade_route(
        self, request: StopTradeRouteRequest
    ) -> List[AnacreonObject]:
        """Stop all trade between 2 planets

        Args:
            request (StopTradeRouteRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def buy_item(self, request: BuyItemRequest) -> List[AnacreonObject]:
        """Buy some quantity of an item from a Mesophon world

        Args:
            request (BuyItemRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def sell_fleet(self, request: SellFleetRequest) -> List[AnacreonObject]:
        """Sell a fleet to a Mesophon world

        Args:
            request (SellFleetRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[AnacreonObject]:  A partial, refreshed version of the
            :py:meth:`~AnacreonAsyncClient.get_objects` response
        """
    async def get_tactical(
        self, battlefield_id: GetTacticalRequest
    ) -> List[Dict[str, Any]]:
        """Get information about the battlefield on a single planet

        This corresponds to the view you get when you zoom in all the way on a
        planet, and you are able to see individual ship groups.

        Args:
            battlefield_id (GetTacticalRequest): Request object containing
                API auth token + details about the operation

        Returns:
            List[Dict[str, Any]]: JSON response
        """
    async def tactical_order(self, order: TacticalOrderRequest) -> bool:
        """Give an order to a ship group orbiting a world

        Args:
            order (TacticalOrderRequest): Request object containing
                API auth token + details about the operation

        Returns:
            bool: ``True`` if your order was carried out successfully
        """
    async def set_history_read(self, history_id: SetHistoryReadRequest) -> bool:
        """Delete a history popup that shows up over planets or fleets

        Args:
            history_id (SetHistoryReadRequest): Request object containing
                API auth token + details about the operation

        Returns:
            bool: ``True`` if your order was carried out successfully
        """
    async def send_message(self, message: SendMessageRequest) -> None:
        """Send a message to another empire

        Args:
            message (SendMessageRequest): Request object containing
                API auth token + details about the operation
        """



import base64
import json
from functools import cached_property
from typing import Tuple, Any
import typing

from pydantic import ConfigDict, BaseModel, validator, PlainSerializer, BeforeValidator


def _snake_case_to_lower_camel(snake: str) -> str:
    """Converts a string that is in `snake_case_form` to `lowerCamelCase`, but special-casing the string "ID"."""

    def ensure_correct_case(pair: Tuple[int, str]) -> str:
        i, word = pair
        if i == 0:
            return word.lower()
        elif word.lower() == "id":
            return word.upper()
        else:
            return word.capitalize()

    camel_words = map(ensure_correct_case, enumerate(snake.split("_")))
    return "".join(camel_words)


_AEON_IP_INTEGER_SENTINEL_VALUE = "AEON2011:ipInteger:v1"
_AEON_IP_INT_POSITIVE = b"IP1+"
_AEON_IP_INT_NEGATIVE = b"IP1-"


def _convert_int_to_aeon_ipinteger(o: Any) -> Any:
    """Convert some value to a Hexarc ``ipInteger`` if appropriate

    The conversion is only done when
    - The parameter ``o`` is a :py:class:`int`
    - The value of ``o`` cannot fit into a 32-bit signed integer

    Args:
        o (Any): Any value

    Returns:
        Any: The value of ``o``, unless it was an int that did not fit into 32
        bits, in which case it is converted to an ``ipInteger``, which is a list of
        the form ``["AEON2011:ipInteger:v1", "KzFQSQQAAACDIVYA"]``
    """
    if isinstance(o, int) and (abs(o) > (2**31 - 1)):
        is_pos = int(o >= 0)
        try:
            num_bytes = abs(o).to_bytes(6, byteorder="big")
        except OverflowError:
            num_bytes = abs(o).to_bytes(1023, byteorder="big")

        size_bytes = (1024).to_bytes(2, byteorder="big")
        sign_bytes = _AEON_IP_INT_POSITIVE if is_pos else _AEON_IP_INT_NEGATIVE

        num_together = sign_bytes[::-1] + size_bytes + num_bytes
        encoded_num = base64.b64encode(num_together)
        return [_AEON_IP_INTEGER_SENTINEL_VALUE, encoded_num]

    return o


def _convert_aeon_ipinteger_to_int(v: Any) -> Any:
    """Convert a Hexarc ``ipInteger`` of the form
    ``["AEON2011:ipInteger:v1", "KzFQSQQAAACDIVYA"]`` into a numeric value

    Encoding a very large number (on the order of billions) in JSON can be
    sketchy because all numbers in JSON are technically 64-bit floating point
    numbers, so after a certain point, not all integers can be represented.

    To solve this problem, Hexarc (the platform on which the Anacreon API runs)
    will encode numbers that cannot fit in a 32 bit signed integer (i.e the
    number has a magnitude beyond 2^31) in a special format. In the JSON, this
    comes across as a 2-element tuple: the first element is a sentinel value to
    check that the JSON list is representing a large number, and the second
    element is the actual encoded number.

    This validator decodes any numbers the Anacreon API encoded in order to
    avoid overflow, and converts them to a python `int`, which should not
    normally overflow.

    Args:
        v (Any): Any value returned by the Anacreon API

    Returns:
        Any: If ``v`` was a Hexarc ``ipInteger``, then the decoded integer value
        is returned. Otherwise, ``v`` is returned.
    """
    if (
        isinstance(v, list)
        and len(v) == 2
        and v[0] == _AEON_IP_INTEGER_SENTINEL_VALUE
        and isinstance(v[1], str)
    ):
        b64_encoded_num: bytes = bytes(v[1], encoding="utf-8")

        # num still has aux data
        encoded_num: bytes = base64.b64decode(b64_encoded_num)

        # decode sign
        encoded_sign: bytes = encoded_num[0:4][::-1]
        sign: int
        if encoded_sign == _AEON_IP_INT_POSITIVE:
            sign = 1
        elif encoded_sign == _AEON_IP_INT_NEGATIVE:
            sign = -1
        else:
            return v

        # bytes [4:6] are the size of the int, we don't care about that in python
        # bytes [6:] are the actual bytes of the integer

        decoded_num: int = int.from_bytes(
            encoded_num[6:], byteorder="big", signed=False
        )
        return decoded_num * sign

    return v


#: Helper to convert large integers from/into the AEON IP integer format.
LargeInt = typing.Annotated[
    int,
    PlainSerializer(_convert_int_to_aeon_ipinteger, when_used="json-unless-none"),
    BeforeValidator(_convert_aeon_ipinteger_to_int),
]


class DeserializableDataclass(BaseModel, metaclass=type):
    # TODO[pydantic]: The following keys were removed: `json_dumps`.
    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-config for more information.
    model_config = ConfigDict(
        alias_generator=_snake_case_to_lower_camel,
        ignored_types=(cached_property, set),
    )



from enum import Enum
from typing import Dict, List, Any, Optional, Union, Type

import uplink
from anacreonlib.exceptions import HexArcException

from anacreonlib.types._deser_utils import DeserializableDataclass
from anacreonlib.types.response_datatypes import Sovereign, ReigningSovereign
from anacreonlib.types.type_hints import Location


class Category(str, Enum):
    BUREAU_OF_TRADE = "bureauOfTrade"
    COMMODITY = "commodity"
    CULTURE = "culture"
    DESIGNATION = "designation"
    DOCTRINE = "doctrine"
    FEATURE = "feature"
    FIXED_UNIT = "fixedUnit"
    GROUND_UNIT = "groundUnit"
    IMPROVEMENT = "improvement"
    INDUSTRY = "industry"
    LAM_UNIT = "LAMUnit"
    MANEUVERING_UNIT = "maneuveringUnit"
    ORBITAL_UNIT = "orbitalUnit"
    REBELLION = "rebellion"
    WORLD_CLASS = "worldClass"


class Role(str, Enum):
    TECH_ADVANCE = "techAdvance"
    SECTOR_CAPITAL = "sectorCapital"
    ORBITAL_DEFENSE_INDUSTRY = "orbitalDefenseIndustry"
    IMPERIAL_CAPITAL = "imperialCapital"
    SHIPYARD_INDUSTRY = "shipyardIndustry"
    SPACEPORT = "spaceport"
    LIFE_SUPPORT = "lifeSupport"
    CITADEL_INDUSTRY = "citadelIndustry"
    ACADEMY = "academy"
    UNIVERSITY = "university"
    ENERGY_INDUSTRY = "energyIndustry"
    CITADEL = "citadel"
    ACADEMY_INDUSTRY = "academyIndustry"
    CONSUMER_GOODS_INDUSTRY = "consumerGoodsIndustry"
    FOUNDATION = "foundation"
    ADMINISTRATION = "administration"
    GROUND_DEFENSE_INDUSTRY = "groundDefenseIndustry"
    RAW_MATERIAL_INDUSTRY = "rawMaterialIndustry"
    SHIPYARD = "shipyard"
    TRADING_HUB = "tradingHub"
    COMPONENT_INDUSTRY = "componentIndustry"


class ScenarioInfoClass(str, Enum):
    CURRENCY_TYPE = "currencyType"
    IMAGE = "image"
    REGION_TYPE = "regionType"
    RESOURCE_TYPE = "resourceType"
    SCENARIO = "scenario"
    SOVEREIGN_TYPE = "sovereignType"
    TRAIT = "trait"


class UserInfo(DeserializableDataclass):
    capital_obj_id: int
    game_id: str
    map_bookmarks: List[Location]
    sovereign_id: int
    ui_options: Any = None
    username: str


class ScenarioInfoElement(DeserializableDataclass):
    is_cargo: Optional[bool] = None
    player_alloc: Optional[bool] = None
    player_product_alloc: Optional[bool] = None
    image_label: Optional[int] = None
    hidden: Optional[bool] = None
    designation_only: Optional[bool] = None
    npe_only: Optional[bool] = None
    is_jump_beacon: Optional[bool] = None
    can_land: Optional[bool] = None
    category: Optional[Category] = None
    scenario_info_class: Optional[ScenarioInfoClass] = None
    id: Optional[int] = None
    image_large: Optional[List[int]] = None
    image_small: Optional[List[int]] = None
    mass: Optional[float] = None
    name_desc: Optional[str] = None
    unid: Optional[str] = None
    attack_value: Optional[int] = None
    short_name: Optional[str] = None
    stats: Optional[List[Union[int, str]]] = None
    build_time: Optional[int] = None
    description: Optional[str] = None
    image_medium: Optional[List[int]] = None
    min_tech_level: Optional[int] = None
    role: Optional[Role] = None
    build_upgrade: Optional[List[int]] = None
    inherit_from: Optional[List[int]] = None
    exports: Optional[List[int]] = None
    max_tech_level: Optional[int] = None
    primary_industry: Optional[int] = None
    requirements: Optional[List[int]] = None
    build_exclusions: Optional[List[int]] = None
    tech_level_advance: Optional[int] = None
    background_color: Optional[List[int]] = None
    background_image: Optional[List[int]] = None
    background_image_low_detail: Optional[List[int]] = None
    background_image_tactical: Optional[List[int]] = None
    exclusions: Optional[List[int]] = None
    ftl: Optional[int] = None
    capital_industry: Optional[int] = None
    capital_type: Optional[int] = None
    build_requirements: Optional[List[int]] = None
    cargo_space: Optional[int] = None
    map_feature_size: Optional[int] = None
    map_size: Optional[List[int]] = None
    name: Optional[str] = None


class ScenarioInfo(DeserializableDataclass):
    scenario_info: List[ScenarioInfoElement]
    sovereigns: List[Union[ReigningSovereign, Sovereign]]
    user_info: UserInfo

    def find_by_unid(self, unid: str) -> ScenarioInfoElement:
        try:
            return next(item for item in self.scenario_info if item.unid == unid)
        except StopIteration:
            raise LookupError(f"Could not find ScenarioInfoElement with unid {unid}")


@uplink.loads.from_json(ScenarioInfo)
def convert_json_to_scenario_info(
    cls: Type[ScenarioInfo], json: Union[Dict[Any, Any], List[Any]]
) -> ScenarioInfo:
    if (
        isinstance(json, list)
        and len(json) == 4
        and all(isinstance(val, str) for val in json)
    ):
        raise HexArcException(json)

    return ScenarioInfo.model_validate(json)



"""This package contains all of the Pydantic models that are used to interact
with the Anacreon API

Because all of the types are Pydantic models, they come with handy methods that
can serialize/deserialize to JSON. You can read more about Pydantic on their
`documentation website <https://pydantic-docs.helpmanual.io/>`_.
"""



"""Utility function to help with decode information from the Anacreon API

Where possible, function signatures have been ordered with partial application 
(see :py:func:`functools.partial`) in mind.
"""

import math
from typing import (
    List,
    Dict,
    Optional,
    Sequence,
    TypeVar,
    Union,
    Tuple,
    cast,
)

from anacreonlib.types.response_datatypes import World, Trait
from anacreonlib.types.scenario_info_datatypes import ScenarioInfoElement
from anacreonlib.types.type_hints import Location

T = TypeVar("T")
U = TypeVar("U")


def flat_list_to_tuples(lst: Sequence[T]) -> List[Tuple[T, T]]:
    """Turn a flat list of alternating key-value pairs into a list of tuples

    Example:
        A common use case of this function is to iterate over a ``resources`` list

        >>> resources = [130, 500, 159, 15000]
        >>> for res_id, res_qty in utils.flat_list_to_tuples(resources):
        ...     print(f"We have {res_qty} of resource id {res_id}")
        ...
        We have 500 of resource id 130
        We have 15000 of resource id 159

    Args:
        lst (Sequence[T]): A flat sequence of alternating key-value pairs

    Returns:
        List[Tuple[T, T]]: A list of key-value tuples
    """
    return list(zip(lst[::2], lst[1::2]))


def flat_list_to_n_tuples(n: int, lst: List[T]) -> List[Tuple[T, ...]]:
    """Converts a list ``[1,2,3,4,5,6,...]`` into a list of tuples ``[(1,2,3), (4,5,6), ...]``
    where the length of each tuple is specified by the parameter ``n``

    Example:
        One use case of this function is to handle import data

        >>> imports = [130, 100, 5700, None, 260, 100, 6400, 500]
        >>> for res_id, pct, optimal, actual in flat_list_to_n_tuples(4, imports):
        ...     print(f"Imported {actual or optimal} units of resource id {res_id} (optimal: {optimal})")
        ...
        Imported 5700 units of resource id 130 (optimal: 5700)
        Imported 500 units of resource id 260 (optimal: 6400)

    Args:
        n (int): The length of each tuple in the returned list
        lst (List[T]): A flat list of all of the items

    Returns:
        List[Tuple[T, ...]]: A list of tuples, where each item in the original list appears exactly once.
    """
    zip_args = [lst[i::n] for i in range(n)]
    return list(zip(*zip_args))


def dist(pointA: Location, pointB: Location) -> float:
    """Retuns the distance between 2 points

    Example:

        >>> pointA = (0, 0)
        >>> pointB = (3, 4)
        >>> utils.dist(pointA, pointB)
        5.0

    Args:
        pointA (Location): An ``(x, y)`` coordinate pair
        pointB (Location): An ``(x, y)`` coordinate pair

    Returns:
        float: The distance between the two points
    """
    ax, ay = pointA
    bx, by = pointB

    dx = bx - ax
    dy = by - ay

    return math.sqrt(dx * dx + dy * dy)


def world_has_trait(
    scninfo: List[ScenarioInfoElement],
    world: World,
    target_trait_id: int,
    include_world_characteristics: bool = True,
) -> bool:
    """Checks whether a world has a given trait, or a child trait extends the
    given trait

    Args:
        scninfo (List[ScenarioInfoElement]): The scenario info for the game.
        world (World): The world to check
        target_trait_id (int): The trait to check the presence of
        include_world_characteristics (bool, optional): Whether or not we should
            take intrinsic world characteristics (e.g world type, designation)
            into account. Defaults to True.

    Returns:
        bool: Whether or not the planet has the given trait.
    """
    trait_dict = world.squashed_trait_dict

    trait_id: int
    for trait_id in trait_dict.keys():
        if target_trait_id == trait_id:
            return True
        elif trait_inherits_from_trait(scninfo, trait_id, target_trait_id):
            return True

    if include_world_characteristics:
        characteristic_traits = (world.world_class, world.designation, world.culture)
        characteristic_inherits_from_target = any(
            trait_inherits_from_trait(scninfo, characteristic, target_trait_id)
            for characteristic in characteristic_traits
        )

        if (
            target_trait_id in characteristic_traits
            or characteristic_inherits_from_target
        ):
            return True

    return False


def trait_inherits_from_trait(
    scninfo: List[ScenarioInfoElement],
    child_trait: Union[Trait, int],
    parent_trait: Union[Trait, int],
) -> bool:
    """Check if a trait inherits from another trait

    Traits can extend other traits. For example, the trait for having abundant
    chronimium deposits, ID ``50``, inherits from the trait for having any
    chronimium deposits, ID ``56``. A world that has trait ID ``50`` also has trait ID
    ``56``. However, this is only implied by the inheritance hierarchy of traits --
    the world object as returned by the API does would only give the "most
    specific" trait. In this case, it would only `explicitly` say the world has
    trait ID ``50``. This function can be used to figure out that the world also has
    trait ID ``56`` as a result.

    Args:
        scninfo (List[ScenarioInfoElement]): The scenario info
        child_trait (Union[Trait, int]): The 'more advanced/specific' trait
            (e.g the trait id for abundant chronimium deposits)
        parent_trait (Union[Trait, int]): The 'less advanced/specific' trait
            (e.g the trait id for chronimium deposits)

    Raises:
        LookupError: Raised if the child trait ID could not be found in the
        scenario info

    Returns:
        bool: ``True`` if the child trait supplants the parent trait, like how a
            sealed arcology supplants a domed city. ``False`` otherwise.
    """
    try:
        # looks up up trait in scenario info based on id. information from Trait object is not used.
        childs_parent_list = next(
            trait.inherit_from for trait in scninfo if trait.id == child_trait
        )
        if childs_parent_list is None:
            return False

        return parent_trait in childs_parent_list or any(
            trait_inherits_from_trait(scninfo, trait, parent_trait)
            for trait in childs_parent_list
        )
    except StopIteration as e:
        raise LookupError("child trait was not found in scenario info") from e
    except KeyError:
        return False


def trait_under_construction(
    squashed_trait_dict: Dict[int, Union[int, Trait]], trait_id: int
) -> bool:
    """Check whether a world is still building a trait

    Args:
        squashed_trait_dict (Dict[int, Union[int, Trait]]): A dict mapping from
            trait ID to either the trait ID or trait object. This can be
            obtained from
            :func:`anacreonlib.types.response_datatypes.World.squashed_trait_dict`

        trait_id (int): The ID of the trait to check for

    Returns:
        bool: ``True`` if the world has the trait, and it is still building.
        ``False`` if the world has fully built the trait, or if it does not have
        the trait at all.
    """
    if trait_id not in squashed_trait_dict.keys():
        return False  # world doesn't have it at all
    trait = squashed_trait_dict[trait_id]
    if isinstance(trait, int):
        return False  # its a simple structure, world has it, and its built
    if trait.build_complete is not None:
        return True

    return False


def world_has_fully_built_trait(
    scninfo: List[ScenarioInfoElement], world: World, target_trait_id: int
) -> bool:
    """Check that a world has fully built a trait to completion.

    This function is aware of the trait inheritance hierarchy. So, for example,
    we would consider a planet that has a sealed arcology (trait ID ``218``) to have
    'fully built' a domed city (trait ID ``88``), because a sealed arcology upgrades
    a domed city.

    Args:
        scninfo (List[ScenarioInfoElement]): The scenario info for the game
        world (World): The world to check for
        target_trait_id (int): The trait ID to check for

    Returns:
        bool: ``True`` if the world has the specified trait, or a successor to
        the specified trait. ``False`` otherwise.
    """
    return world_has_trait(
        scninfo, world, target_trait_id, True
    ) and not trait_under_construction(world.squashed_trait_dict, target_trait_id)


def does_trait_depend_on_trait(
    scninfo: List[ScenarioInfoElement],
    trait_a: int,
    trait_b: int,
) -> bool:
    """Check if a ``trait_b`` has another ``trait_a`` anywhere in the chain of
    requirements.

    For example, a sealed arcology (trait ID 218) is an upgrade to a domed city
    (trait id 88). So, the call
    ``utils.does_trait_depend_on_trait(scninfo, 218, 88)`` would return ``True``
    because trait ID 218 needs trait ID 88 to be built first.

    Note that there can be multiple upgrade paths to a more advanced
    trait/improvement. For example, the the sealed arcology ruins (ID 219) are
    also a predecessor to the sealed arcology (ID 218).

    Args:
        scninfo (List[ScenarioInfoElement]): The scenario info
        trait_a (int): The ID of the 'more advanced' trait
        trait_b (int): The ID of the 'less advanced' trait

    Returns:
        bool: Whether ``trait_a`` needs ``trait_b`` to be built first.
    """
    more_advanced_improvement = scninfo[trait_a]
    if more_advanced_improvement.build_upgrade is not None:
        return trait_b in more_advanced_improvement.build_upgrade or any(
            does_trait_depend_on_trait(scninfo, trait, trait_b)
            for trait in more_advanced_improvement.build_upgrade
        )
    return False


def get_world_primary_industry_products(world: World) -> Optional[List[int]]:
    """Get the list of resource IDs that a planet produces with its primary
    industry

    Args:
        world (World): The world in question

    Returns:
        Optional[List[int]]: A list of resource IDs that the planet's
        designation produces. If this is ``None``, then the planet's designation
        does not produce items (e.g it may be a foundation world, or autonomous)
    """
    primary_industry = next(
        (
            trait
            for trait in world.traits
            if isinstance(trait, Trait) and trait.is_primary
        ),
        None,
    )

    if primary_industry is None or primary_industry.production_data is None:
        return None

    return [
        cast(int, res_id)
        for res_id, pct_alloc, cannot_build in flat_list_to_n_tuples(
            3, primary_industry.build_data
        )
    ]



from enum import Enum
from typing import Literal, Tuple

TechLevel = Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

Location = Tuple[float, float]
"""
A location is a tuple of 2 numbers
    - x pos
    - y pos
"""

Circle = Tuple[float, float, float]
"""
A Circle is a tuple of 3 numbers: 
    - x pos of center
    - y pos of center
    - radius, in lightyears
"""

Arc = Tuple[float, float, float, float, float]
"""
An Arc is a tuple of 5 numbers:
    - x pos of center
    - y pos of center
    - radius, in light years
    - starting angle in radians
    - ending angle in radians
"""


class BattleObjective(str, Enum):
    INVASION = "invasion"
    SPACE_SUPREMACY = "spaceSupremacy"
    REINFORCE_SIEGE = "reinforceSiege"


class SiegeStatus(str, Enum):
    ATTACK_FAILING = "attackFailing"
    ATTACK_WINNING = "attackWinning"
    DEFENSE_FALIING = "defenseFailing"
    DEFENSE_WINNING = "defenseWinning"



import setuptools
from setuptools import setup

with open("README.rst", "r") as f:
    long_desc_contents = f.read()

with open("VERSION", "r") as f:
    version = f.read().strip()

setup(
    name="anacreonlib",
    version=version,
    description="""This library provides a Python interface to the API of Anacreon 3, which is an online 4X game 
      produced by Kronosaur Productions, LLC.""",
    long_description=long_desc_contents,
    long_description_content_type="text/markdown",
    url="https://github.com/ritikmishra/anacreonlib",
    author="Ritik Mishra",
    author_email="ritik.mishra314@gmail.com",
    license="MIT",
    packages=setuptools.find_packages(),
    package_data={"anacreonlib": ["py.typed", "anacreon_async_client.pyi"]},
    install_requires=["uplink[aiohttp]", "pydantic"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        # "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        # "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.8",
    test_suite="tests",
)



import json
from enum import Enum
from typing import Any, Dict, List, Type, TypeVar, Union, Optional, cast

from pydantic import ConfigDict, BaseModel, Field
from uplink import dumps

from anacreonlib.types._deser_utils import _snake_case_to_lower_camel, LargeInt
from anacreonlib.types.type_hints import BattleObjective


class SerializableDataclass(BaseModel):
    """Base class for a serializable Anacreon request object. Contains common Pydantic config settings."""

    model_config = ConfigDict(
        populate_by_name=True, alias_generator=_snake_case_to_lower_camel
    )


class AuthenticationRequest(SerializableDataclass):
    username: str
    password: str
    actual: bool = True


class AnacreonApiRequest(SerializableDataclass):
    """Base class for request bodies to most of the endpoints for the Anacreon API"""

    auth_token: str
    game_id: str
    sovereign_id: LargeInt
    sequence: Optional[LargeInt] = None


class TacticalOrderType(str, Enum):
    ORBIT = "orbit"
    LAND = "land"
    TARGET = "target"


class BattlePlan(SerializableDataclass):
    battlefield_id: int = Field(..., alias="battleFieldID")
    objective: BattleObjective
    enemy_sovereign_ids: List[int] = Field(alias="enemySovereignIDs")


# region ACTION TYPES
class DeployFleet(SerializableDataclass):
    source_obj_id: int
    resources: List[LargeInt]


class TransferFleet(SerializableDataclass):
    fleet_obj_id: int
    dest_obj_id: int
    resources: List[LargeInt]
    # source_obj_id: None = Field(None)


class DisbandFleet(SerializableDataclass):
    fleet_obj_id: int
    dest_obj_id: int


class RenameObject(SerializableDataclass):
    obj_id: int
    name: str


class SetFleetDestination(SerializableDataclass):
    obj_id: int
    dest: int


class Attack(SerializableDataclass):
    attacker_obj_id: int
    battle_plan: BattlePlan


class AbortAttack(SerializableDataclass):
    battlefield_id: int = Field(..., alias="battleFieldID")


class LaunchJumpMissile(SerializableDataclass):
    source_obj_id: int = Field(..., alias="objID")
    target_obj_id: int


class DesignateWorld(SerializableDataclass):
    source_obj_id: int
    new_designation: int


class AlterImprovement(SerializableDataclass):
    """Used to both build and destroy improvements"""

    source_obj_id: int
    improvement_id: int


class SetIndustryAlloc(SerializableDataclass):
    world_id: int = Field(..., alias="objID")
    industry_id: int
    alloc_value: Union[int, float] = Field(..., ge=0, le=100)


class SetProductAlloc(SerializableDataclass):
    world_id: int = Field(..., alias="objID")
    industry_id: int
    alloc: List[Union[int, float]]


class TradeRouteTypes(str, Enum):
    TECH = "tech"
    CONSUMPTION = "consumption"
    DEFAULT = "addDefaultRoute"
    ADD_EXPORT_ROUTE = "addExportRoute"
    SET_EXPORT_QUOTA = "setExportQuota"


class SetTradeRoute(SerializableDataclass):
    importer_id: int = Field(..., alias="objID")
    exporter_id: int = Field(..., alias="sourceObjID")
    alloc_type: Union[TradeRouteTypes, str]
    alloc_value: Optional[Union[str, float]] = None
    res_type_id: Optional[int] = Field(None, alias="resType")


class StopTradeRoute(SerializableDataclass):
    planet_id_a: int = Field(..., alias="objID")
    planet_id_b: int = Field(..., alias="sourceObjID")


class BuyItem(SerializableDataclass):
    source_obj_id: int
    item_id: int
    item_count: LargeInt


class TacticalOrder(SerializableDataclass):
    battlefield_id: int = Field(..., alias="objID")
    order: Union[TacticalOrderType, str]
    squadron_id: int = Field(..., alias="tacticalID")
    orbit: Optional[float] = None
    tactical_id: Optional[int] = None


class SetHistoryRead(SerializableDataclass):
    history_id: LargeInt


# endregion


# region API TYPES
class SellFleet(SerializableDataclass):
    fleet_id: int = Field(..., alias="objID")
    buyer_obj_id: int
    resources: List[LargeInt]


class SendMessage(SerializableDataclass):
    recipient_id: int
    message_text: str


class GetTactical(SerializableDataclass):
    battlefield_id: int = Field(..., alias="objID")


class DeployFleetRequest(DeployFleet, AnacreonApiRequest):
    pass


class TransferFleetRequest(TransferFleet, AnacreonApiRequest):
    pass


class DisbandFleetRequest(DisbandFleet, AnacreonApiRequest):
    pass


class RenameObjectRequest(RenameObject, AnacreonApiRequest):
    pass


class SetFleetDestinationRequest(SetFleetDestination, AnacreonApiRequest):
    pass


class AttackRequest(Attack, AnacreonApiRequest):
    pass


class AbortAttackRequest(AbortAttack, AnacreonApiRequest):
    pass


class LaunchJumpMissileRequest(LaunchJumpMissile, AnacreonApiRequest):
    pass


class DesignateWorldRequest(DesignateWorld, AnacreonApiRequest):
    pass


class AlterImprovementRequest(AlterImprovement, AnacreonApiRequest):
    pass


class SetIndustryAllocRequest(SetIndustryAlloc, AnacreonApiRequest):
    pass


class SetProductAllocRequest(SetProductAlloc, AnacreonApiRequest):
    pass


class SetTradeRouteRequest(SetTradeRoute, AnacreonApiRequest):
    pass


class StopTradeRouteRequest(StopTradeRoute, AnacreonApiRequest):
    pass


class BuyItemRequest(BuyItem, AnacreonApiRequest):
    pass


class SellFleetRequest(SellFleet, AnacreonApiRequest):
    pass


class GetTacticalRequest(GetTactical, AnacreonApiRequest):
    pass


class TacticalOrderRequest(TacticalOrder, AnacreonApiRequest):
    pass


class SetHistoryReadRequest(SetHistoryRead, AnacreonApiRequest):
    pass


class SendMessageRequest(SendMessage, AnacreonApiRequest):
    pass


# endregion


@dumps.to_json(AnacreonApiRequest)
def pydantic_request_converter(
    cls: type, inst: SerializableDataclass
) -> Dict[Any, Any]:
    """
    Converts request models to dict, with field aliases as keys

    The default pydantic converter in uplink doesn't use aliases, which we use extensively.
    """
    return cast(Dict[Any, Any], json.loads(inst.model_dump_json(by_alias=True)))



from enum import Enum
import functools
from contextlib import suppress
from typing import (
    Annotated,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import uplink

from anacreonlib.exceptions import HexArcException
from anacreonlib.types._deser_utils import DeserializableDataclass, LargeInt
from anacreonlib.types.type_hints import (
    Arc,
    BattleObjective,
    Circle,
    Location,
    SiegeStatus,
    TechLevel,
)
from pydantic import AliasGenerator, ConfigDict, Field, ValidationError, model_validator

__all__ = (
    "AuthenticationResponse",
    "AnacreonObject",
    "AnacreonObjectWithId",
    "News",
    "SovereignActions",
    "SovereignRelationship",
    "ExplorationGrid",
    "SovereignStats",
    "MesophonTrait",
    "BattlePlanDetails",
    "RegionShape",
    "Trait",
    "Rebellion",
    "Siege",
    "HistoryElement",
    "History",
    "TradeRoute",
    "RevIndex",
    "NebulaType",
    "World",
    "OwnedWorld",
    "Sovereign",
    "ReigningSovereign",
    "OwnSovereign",
    "BattlePlanObject",
    "Fleet",
    "DestroyedSpaceObject",
    "UpdateObject",
    "RegionObject",
    "Relationship",
    "Selection",
)

# response datatype parent/abstract classes


class AuthenticationResponse(DeserializableDataclass):
    auth_token: str
    rights: List[str]
    scoped_credentials: LargeInt
    username: str


T = TypeVar("T")
ObjectClass = Annotated[T, Field(alias="class")]


class AnacreonObject(DeserializableDataclass):
    object_class: ObjectClass[str]


class AnacreonObjectWithId(AnacreonObject):
    """Not all anacreon objects have an ID, most notably, UpdateObject"""

    id: int


# region: subclasses


class News(DeserializableDataclass):
    subject: int
    text: str


class SovereignActions(DeserializableDataclass):
    attacks_initiated: int
    offensives_initiated: int
    worlds_conquered: int


class SovereignRelationship(DeserializableDataclass):
    first_contact: int
    our_actions: SovereignActions
    their_actions: SovereignActions


class ExplorationGrid(DeserializableDataclass):
    radius: float
    explored_outline: List[List[float]]


class SovereignStats(DeserializableDataclass):
    fleets: LargeInt
    population: int
    resources: Optional[List[LargeInt]] = None
    tech_level: TechLevel
    worlds: int


class MesophonTrait(DeserializableDataclass):
    """Trait applied to the Mesophon sovereign (NPE trading empire that players can buy ships from)"""

    #: A list which alternates between resource ID and the number of aes this
    #: empire is willing to pay per unit (they are the buyer)
    buy_prices: List[Union[int, float]]

    #: A list which alternates between resource ID and the number
    #: of aes this empire is willing to accept per unit (they are the seller)
    sell_prices: List[Union[int, float]]
    trait_id: int


class BattlePlanDetails(DeserializableDataclass):
    enemy_sovereign_ids: Optional[List[int]] = Field(None, alias="enemySovereignIDs")
    objective: BattleObjective
    sovereign_id: int
    status: str

    @model_validator(mode="after")
    def validate_enemy_sovereign_ids(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if (
            values.get("enemy_sovereign_ids") is None
            and values.get("objective") != BattleObjective.REINFORCE_SIEGE
        ):
            raise ValueError(
                "'enemy_sovereign_ids' can only be None when 'objective' is BattleObjective.REINFORCE_SIEGE"
            )
        return values


class RegionShape(DeserializableDataclass):
    holes: Optional[List[List[float]]] = None
    outline: List[float]


class Trait(DeserializableDataclass):
    allocation: float
    build_data: List[Union[float, None, List[Any]]]
    #: ``True``` if this structure is the primary industry on the world
    #: (i.e it belongs to the designation)
    is_primary: Optional[bool] = None
    production_data: Optional[List[Union[float, None]]] = None
    is_fixed: Optional[bool] = None
    target_allocation: float
    trait_id: int
    build_complete: Optional[int] = None
    work_units: float


class Rebellion(DeserializableDataclass):
    #: if popular support is greater than ``0```, it indicates rebel support
    #: if it is less than ``0```, it indicates imperial support
    popular_support: float
    rebel_forces: float
    rebellion_start: int
    trait_id: int


class Siege(AnacreonObjectWithId):
    object_class: ObjectClass[Literal["siege"]]
    anchor_obj_id: int
    attack_forces: float
    defense_forces: float
    name: str
    news: Optional[List[News]] = None
    pos: Location
    resources: Optional[List[float]] = None
    sovereign_id: int
    status: Optional[SiegeStatus] = None
    timeLeft: Optional[int] = None


class HistoryElement(DeserializableDataclass):
    #: ID of this history element, which can be passed to the setHistoryRead endpoint to clear this message
    id: int

    #: ID of the object this message applies to
    obj_id: int
    subject: int
    text: str


class History(AnacreonObject):
    object_class: ObjectClass[Literal["history"]]
    history: List[HistoryElement]


class TradeRoute(DeserializableDataclass):
    """
    :ivar import_tech: tuple of the desired tech level to acheive, and how many levels uplifted the planet actually is
    :ivar reciprocal: if true, the data for this trade route is attached to the partner object.
    """

    imports: Optional[List[Union[float, None]]] = None
    exports: Optional[List[Union[float, None]]] = None

    #: Third tuple element is usually not present. If present, it indicates why
    #: the world cannot be uplifted to the desired tech level.
    import_tech: Union[Tuple[int, int], Tuple[int, int, Any], None] = None
    export_tech: Union[Tuple[int, int], Tuple[int, int, Any], None] = None
    partner_obj_id: int
    reciprocal: Optional[bool] = Field(None, alias="return")


# endregion
# region: anacreon objects


class RevIndex(str, Enum):
    """Enum of social orders a world can have"""

    HAPPY = "happy"
    CONTENT = "content"
    DISSATISFIED = "dissatisfied"
    AGGRIEVED = "aggrieved"
    RIOTING = "rioting"
    REBELLING = "rebelling"
    CIVIL_WAR = "civil war"


class NebulaType(int, Enum):
    """Enum of types of space regions a world can be in

    (technically, a world cannot be in a `RIFT_ZONE`)
    (these values are a guess)
    """

    CLEAR_SPACE = 1
    LIGHT_NEBULA = 2
    DARK_NEBULA = 3
    RIFT_ZONE = 4


class World(AnacreonObjectWithId):
    object_class: ObjectClass[Literal["world"]]

    #: This is a trait ID that represents something intrinsic about the world's
    #: population
    culture: int

    #: This trait ID corresponds to which designation the world has
    designation: int
    efficiency: float = Field(..., ge=0, le=100)
    name: str

    #: This is a list of IDs of fleets that are stationed at this world
    near_obj_ids: Optional[List[int]] = Field(None, alias="nearObjIDs")
    orbit: List[float]
    population: int
    pos: Location

    #: This is a list alternating between resource ID and quantity stockpiled
    #: on the world
    resources: Optional[List[LargeInt]] = None
    sovereign_id: int
    tech_level: TechLevel
    traits: List[Union[int, Trait, Rebellion]]
    world_class: int
    trade_routes: Optional[List[TradeRoute]] = None

    rev_index: Optional[RevIndex] = None

    battle_plan: Optional[BattlePlanDetails] = None

    #: If the world is going to change tech levels soon, the value of this
    #: field is the target tech level
    target_tech_level: Optional[TechLevel] = None

    #: If the population is going to change, the value of this field is
    #: what the planet is heading towards
    #:
    #: Unit is millions of people
    target_population: Optional[int] = None

    region: NebulaType = NebulaType.CLEAR_SPACE

    @functools.cached_property
    def trade_route_partners(self) -> Optional[Dict[int, TradeRoute]]:
        """Returns a set of all of the trade route partners of this world"""
        if self.trade_routes is not None:
            return {
                trade_route.partner_obj_id: trade_route
                for trade_route in self.trade_routes
            }
        return None

    @functools.cached_property
    def squashed_trait_dict(self) -> Dict[int, Union[int, Trait]]:
        """Return a dict mapping from trait ID to either trait ID or trait object"""
        trait_dict: Dict[int, Union[int, Trait]] = {}
        for trait in self.traits:
            if isinstance(trait, int):
                trait_dict[trait] = trait
            elif isinstance(trait, Trait):
                trait_dict[trait.trait_id] = trait

        return trait_dict

    @functools.cached_property
    def resource_dict(self) -> Dict[int, float]:
        """A dict mapping from resource ID to resource qty on the world"""
        if self.resources is not None:
            return dict(zip(self.resources[::2], self.resources[1::2]))
        else:
            return dict()


class OwnedWorld(World):
    """This is a world we know belongs to the current user because you are
    guaranteed to see certain fields
    """

    base_consumption: List[Union[int, None]]
    news: Optional[List[News]] = None

    trade_route_max: Optional[int] = None

    rev_index: RevIndex


class Sovereign(AnacreonObjectWithId):
    """Any sovereign that has ever played in the current game"""

    object_class: ObjectClass[Literal["sovereign"]]

    #: A measure of the empire's strength relative to you
    imperial_might: int
    name: str
    relationship: Optional[SovereignRelationship] = None

    # for some reason, dead sovereigns can have a doctrine
    # , and alive sovereigns might not have a doctrine id
    doctrine: Optional[int] = None

    traits: Optional[List[MesophonTrait]] = None


class ReigningSovereign(Sovereign):
    """Some Sovereigns are abdicated. This class is only for sovereigns currently playing the game."""

    capital_id: int

    stats: SovereignStats

    # Shown if you find their capital
    founded_on: Optional[int] = None
    territory: Optional[List[Union[Circle, Arc]]] = None


class OwnSovereign(ReigningSovereign):
    """Represents the sovereign belonging to the user who is currently logged in"""

    admin_range: List[Union[Circle, Arc]]
    exploration_grid: ExplorationGrid

    #: Alternating list between resource ID and resource quantity
    #: As long as is only one currency (aes), this list will only have 2
    #: elements max
    funds: List[Union[int, float]]  # todo: determine type
    secession_chance: float
    stats: SovereignStats


class BattlePlanObject(AnacreonObjectWithId):
    object_class: ObjectClass[Literal["battlePlan"]]
    battle_plan: BattlePlanDetails


class Fleet(AnacreonObjectWithId):
    object_class: ObjectClass[Literal["fleet"]]

    #: Whether this is a jumpship fleet, starship/ramship fleet, or explorer
    #: fleet
    ftl_type: str
    name: str
    sovereign_id: int
    resources: List[int]

    #: For fleets, you might get news if they were attacked by jumpmissiles
    news: Optional[List[News]] = None

    #: ``None`` if this fleet is not currently stationed at a world. Otherwise,
    #: this is the ID of the world where this fleet is stationed.
    anchor_obj_id: Optional[int] = None
    battle_plan: Optional[BattlePlanDetails] = None
    pos: Location
    destination: Optional[Location] = Field(None, alias="dest")
    dest_id: Optional[int] = None

    #: If not ``None``, corresponds to the :py:attr:`UpdateObject.update`
    #: on which this fleet will arrive at its destination.
    eta: Optional[int] = None

    region: NebulaType = NebulaType.CLEAR_SPACE


class DestroyedSpaceObject(AnacreonObjectWithId):
    object_class: ObjectClass[Literal["destroyedSpaceObject"]]


class UpdateObject(AnacreonObject):
    object_class: ObjectClass[Literal["update"]]

    #: Number of milliseconds until the next watch update happens on the server
    next_update_time: int

    #: This is a unique ID associated with each state update. Using this number,
    #: the Anacreon server can know exactly what data it needs to send us in
    #: order to catch us up to speed with what changed since we last talked.
    sequence: LargeInt

    #: Number of watches that have occured since the game started
    update: float

    #: The in-game calendar year in which the game started (e.g ``4021``)
    year0: int


class RegionObject(AnacreonObjectWithId):
    """Typically used to encode the location of nebulas, rift zones, and clear space"""

    object_class: ObjectClass[Literal["region"]]
    shape: List[RegionShape]

    #: Corresponds to the :py:class:`ScenarioInfoElement` for this region type
    type: int


class Relationship(AnacreonObjectWithId):
    """Partial update for sovereigns"""

    object_class: ObjectClass[Literal["relationship"]]
    relationship: SovereignRelationship


class Selection(AnacreonObjectWithId):
    """
    The API returns this object if it wants the UI to
    select something as a result of the API request

    For example, when you deploy a fleet, the API returns
    one of these objects in order to signal that the new
    fleet should be selected.
    """

    object_class: ObjectClass[Literal["selection"]]


# endregion
# region: utility functions


def _init_obj_subclasses() -> List[Type[AnacreonObject]]:
    subclasses = AnacreonObject.__subclasses__()
    for subcls in subclasses:
        subclasses.extend(subcls.__subclasses__())
    return subclasses[::-1]


_anacreon_obj_subclasses = _init_obj_subclasses()


@uplink.response_handler
def handle_hexarc_error_response(response: Any) -> Any:
    res_json = response.json()
    if (
        type(res_json) == list
        and len(res_json) == 4
        and all(isinstance(val, str) for val in res_json)
    ):
        raise HexArcException(res_json)
    return response


def _convert_json_to_anacreon_obj(
    cls: Type[DeserializableDataclass], json: Dict[Any, Any]
) -> Any:
    classes_to_try: List[Type[DeserializableDataclass]] = list()
    if cls is AnacreonObject:
        classes_to_try.extend(_anacreon_obj_subclasses)
    else:
        classes_to_try.append(cls)

    with suppress(ValueError):
        classes_to_try.remove(AnacreonObject)
    with suppress(ValueError):
        classes_to_try.remove(AnacreonObjectWithId)

    for subcls in classes_to_try:
        try:
            return subcls.model_validate(json)
        except ValidationError as e:
            pass

    return json


convert_json_to_anacreon_obj = uplink.loads.from_json(AnacreonObject)(
    _convert_json_to_anacreon_obj
)

# endregion



class AuthenticationException(Exception):
    """Thrown when we have issues authenticating"""

    pass


class HexArcException(Exception):
    """Thrown when we try to make a request that violates some invariant of Anacreon

    For example,
        - You cannot sell resources that you don't have
        - You cannot designate a world that doesn't belong to you
    and so on.
    """

    pass
